#include <avr\io.h>
#include <avr\interrupt.h>


#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define DDR_LED1 0b00000110
#define DDR_LED2 0b00001001
#define DDR_LED3 0b00001010
#define DDR_LED4 0b00000011
#define DDR_LED5 0b00000101
#define DDR_LED6 0b00010100
#define DDR_LED7 0b00010001
#define DDR_LED8 0b00011000
#define DDR_LED9 0b00001100
#define DDR_LED0 0b00010010

#define PORT_LED1_ROSU  0b00000100
#define PORT_LED1_VERDE 0b00000010

#define PORT_LED2_ROSU  0b00000001
#define PORT_LED2_VERDE 0b00001000

#define PORT_LED3_ROSU  0b00000010
#define PORT_LED3_VERDE 0b00001000

#define PORT_LED4_ROSU  0b00000001
#define PORT_LED4_VERDE 0b00000010

#define PORT_LED5_ROSU  0b00000001
#define PORT_LED5_VERDE 0b00000100

#define PORT_LED6_ROSU  0b00010000
#define PORT_LED6_VERDE 0b00000100

#define PORT_LED7_ROSU  0b00000001
#define PORT_LED7_VERDE 0b00010000

#define PORT_LED8_ROSU  0b00010000
#define PORT_LED8_VERDE 0b00001000

#define PORT_LED9_ROSU  0b00000100
#define PORT_LED9_VERDE 0b00001000

#define PORT_LED0_ROSU  0b00000010
#define PORT_LED0_VERDE 0b00010000

#define DDR_LED DDRA
#define PORT_LED PORTA

#define INC(x) x = x % 2 + 1
#define INC_NO_REASSIGN(x) x % 2 + 1

typedef struct {
	uint8_t port; /* valoarea ce trebuie atribuita lui PORT_LED pentru a se aprinde */
	uint8_t ddr; /* valoare ce trebuie atribuita lui DDR_LED pentru a se aprinde */
} led_t;

led_t matrice_led[3][3]; /* ledurile matricei x si 0 */
uint8_t m[4][4]; /* matricea starii ledurilor; 1 = rosu, 2 = verde, 0 daca nu e aprins */
led_t led_stare; /* led care indica jucatorul curent */
char culori[][5] = { "", "rosu", "verde" };
uint8_t playerCurent = 1; /* poate fi 1 sau 2; 1 = rosu, 2 = verde */
uint8_t xCurent = 1; uint8_t yCurent = 1; /* pozitia de start e stanga sus */
uint8_t nrLEDuriAprinse = 0;
uint8_t gameEnded = 0;
uint8_t pozitiiCastigatoare[3][2];
uint8_t change;
uint8_t letsBEEP;

/**
 * Functie care initializeaza starea ledurilor in memorie
 */
void initLEDs()
{
	int i, j;
	
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++) {
			matrice_led[i][j].ddr = 0;
			matrice_led[i][j].port = 0;
			m[i+1][j+1] = 0;
		}

	led_stare.ddr = 0;
	led_stare.port = 0;
	m[0][0] = 0;
}

void reset() {
	initLEDs();
	playerCurent = 1;
	xCurent = 1;
	yCurent = 1;
	nrLEDuriAprinse = 0;
	gameEnded = 0;
	change = 0;
	letsBEEP = 0;
}

/**
 * Functie care aprinde ledul selectat
 */
void turnLEDon(int i, int j, const char culoare[], int palpaie) {
	if (i == 0 && j == 0) {
		led_stare.ddr = DDR_LED0;
		led_stare.port = (strcmp(culoare, "rosu") == 0) ? PORT_LED0_ROSU : PORT_LED0_VERDE;
	} else

	if (i == 1 && j == 1) {
		matrice_led[i-1][j-1].ddr = DDR_LED1;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED1_ROSU : PORT_LED1_VERDE;
	} else
	if (i == 1 && j == 2) {
		matrice_led[i-1][j-1].ddr = DDR_LED2;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED2_ROSU : PORT_LED2_VERDE;
	} else
	if (i == 1 && j == 3) {
		matrice_led[i-1][j-1].ddr = DDR_LED3;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED3_ROSU : PORT_LED3_VERDE;
	} else

	if (i == 2 && j == 1) {
		matrice_led[i-1][j-1].ddr = DDR_LED4;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED4_ROSU : PORT_LED4_VERDE;
	} else
	if (i == 2 && j == 2) {
		matrice_led[i-1][j-1].ddr = DDR_LED5;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED5_ROSU : PORT_LED5_VERDE;
	} else
	if (i == 2 && j == 3) {
		matrice_led[i-1][j-1].ddr = DDR_LED6;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED6_ROSU : PORT_LED6_VERDE;
	} else

	if (i == 3 && j == 1) {
		matrice_led[i-1][j-1].ddr = DDR_LED7;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED7_ROSU : PORT_LED7_VERDE;
	} else
	if (i == 3 && j == 2) {
		matrice_led[i-1][j-1].ddr = DDR_LED8;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED8_ROSU : PORT_LED8_VERDE;
	} else
	if (i == 3 && j == 3) {
		matrice_led[i-1][j-1].ddr = DDR_LED9;
		matrice_led[i-1][j-1].port = (strcmp(culoare, "rosu") == 0) ? PORT_LED9_ROSU : PORT_LED9_VERDE;
	}
	if (!palpaie)
		m[i][j] = (strcmp(culoare, "rosu") == 0) ? 1 : 2;
}

/**
 * Functie care stinge ledul selectat
 */
void turnLEDoff(int i, int j, int palpaie) {
	if (i == 0 && j == 0) {
		led_stare.ddr = 0b00000000;
		led_stare.port = 0b00000000;
	} else

	if (i == 1 && j == 1) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 1 && j == 2) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 1 && j == 3) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else

	if (i == 2 && j == 1) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 2 && j == 2) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 2 && j == 3) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else

	if (i == 3 && j == 1) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 3 && j == 2) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	} else
	if (i == 3 && j == 3) {
		matrice_led[i-1][j-1].ddr = 0b00000000;
		matrice_led[i-1][j-1].port = 0b00000000;
	}
	if (!gameEnded && !palpaie)
		m[i][j] = 0;
}

/**
 * Functie care lumineaza led-urile conform starii curente din memorie
 */
void displayLEDs(int delay_ms) {
	int i, j;
	int delay = (delay_ms == 0) ? 2 : delay_ms;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++) {
			DDR_LED = matrice_led[i][j].ddr;
			PORT_LED = matrice_led[i][j].port;
			_delay_ms(delay);
		}

	DDR_LED = led_stare.ddr;
	PORT_LED = led_stare.port;
	_delay_ms(delay);
}

/**
 * Functie care verifica daca cineva a castigat si memoreaza pozitiile castigatoare;
 * pozitiile castigatoare vor fi folosite la clipociri pentru a semnaliza linia castigatoare
 */
int aCastigat() {
	if (m[1][1] == m[1][2] && m[1][2] == m[1][3] && m[1][1] != 0) { // prima linie
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 1;

		pozitiiCastigatoare[1][0] = 1;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 1;
		pozitiiCastigatoare[2][1] = 3;
		return 1;
	}

	if (m[1][1] == m[2][1] && m[2][1] == m[3][1] && m[1][1] != 0) { // prima coloana
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 1;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 1;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 1;
		return 1;
	}

	if (m[2][1] == m[2][2] && m[2][2] == m[2][3] && m[2][1] != 0) { // a doua linie
		pozitiiCastigatoare[0][0] = 2;
		pozitiiCastigatoare[0][1] = 1;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 2;
		pozitiiCastigatoare[2][1] = 3;
		return 1;
	}

	if (m[1][2] == m[2][2] && m[2][2] == m[3][2] && m[1][2] != 0) { // a doua coloana
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 2;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 2;
		return 1;
	}

	if (m[3][1] == m[3][2] && m[3][2] == m[3][3] && m[3][1] != 0) { // a treia linie
		pozitiiCastigatoare[0][0] = 3;
		pozitiiCastigatoare[0][1] = 1;

		pozitiiCastigatoare[1][0] = 3;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 3;
		return 1;
	}

	if (m[1][3] == m[2][3] && m[2][3] == m[3][3] && m[1][3] != 0) { // a treia coloana
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 3;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 3;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 3;
		return 1;
	}

	if (m[1][1] == m[2][2] && m[2][2] == m[3][3] && m[1][1] != 0) { // diagonala principala
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 1;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 3;
		return 1;
	}

	if (m[1][3] == m[2][2] && m[2][2] == m[3][1] && m[1][3] != 0) { // diagonala secundara
		pozitiiCastigatoare[0][0] = 1;
		pozitiiCastigatoare[0][1] = 3;

		pozitiiCastigatoare[1][0] = 2;
		pozitiiCastigatoare[1][1] = 2;

		pozitiiCastigatoare[2][0] = 3;
		pozitiiCastigatoare[2][1] = 1;
		return 1;
	}

	return 0;
}

void initTimer() {
	/* timer pentru clipoceli */
	// CTC : TTCR1B - WGM12 = 1 si setat prescaler la clk/1024
	TCCR1B = 1 << WGM12 | 1 << CS12 | 1 << CS10;

	// intrerupere la o zecime de secunda
	OCR1A = 1562;

	//TIMSK activat OCIE1A
	TIMSK = 1 << OCIE1A;


	/* timer pentru speaker */
	// CTC: TTCR2 - WGM21 = 1 si setat prescaler la clk/1024
	TCCR2 = 1 << WGM21 | 1 << CS22 | 1 << CS21 | 1 << CS20;

	// interupere la o milisecunda
	OCR2 = 15;

	// TIMSK activat OCIE2
	TIMSK |= 1 << OCIE2;

}

void initSpeaker() {
	DDRB |= 1 << 3 | 1; // 0b00001001;
	PORTB &= ~(1 << 3) | 1;
}

ISR(TIMER1_COMPA_vect) {
	int i;

	if (!gameEnded) { /* daca jocul nu s-a terminat, clipocim doar pozitia curenta */
		if (matrice_led[yCurent-1][xCurent-1].port != 0b00000000)
			turnLEDoff(yCurent, xCurent, 1);
		else
			turnLEDon(yCurent, xCurent, culori[playerCurent], 1);
	}
	else { /* daca jocul s-a terminat, clipocim ledurile castigatoare */
		for (i = 0; i < 3; i++) {
			if (matrice_led[pozitiiCastigatoare[i][0]-1][pozitiiCastigatoare[i][1]-1].ddr != 0b00000000) {
				/* daca ledurile sunt aprinse, le stingem */
				turnLEDoff(pozitiiCastigatoare[i][0], pozitiiCastigatoare[i][1], 1);
				turnLEDoff(0, 0, 1);
			}
			else { /* daca ledurile sunt stinse, le aprindem */
				turnLEDon(pozitiiCastigatoare[i][0], pozitiiCastigatoare[i][1], culori[m[pozitiiCastigatoare[i][0]][pozitiiCastigatoare[i][1]]], 0);
				turnLEDon(0, 0, culori[m[0][0]], 0);
			}
		}
	}
}

ISR(TIMER2_COMP_vect) {
	if (letsBEEP) {
		if (!change) { // PORTB.0 == 0;
			change = 1;
			PORTB |= 1;
		}
		else {
			change = 0;
			PORTB &= ~(1);
		}
	}
}

ISR(BADISR_vect) {
	 /* nu vrem sa apara intreruperi aiurea care sa ne reseteze jocul,
		asa ca le indrumam aici si le spunem la revedere */
}

int main() {
	uint8_t mutaUrmatoareaPozitie = 0;
	uint8_t move = 0;
	uint8_t i, j, ok = 0;
	sei();

	initSpeaker();
	initTimer();

	initLEDs();
	DDRC = 0x00;

	turnLEDon(0, 0, "rosu", 0);

	while (1) {
		letsBEEP = 0;
		if (PINC != 0xFF) {
			switch(PINC) {
				case 0b11111110: /* save_button */

					break;

				case 0b11111101: /* load_button */

					break;

				case 0b11111011: /* right_button */
					if (!gameEnded) {
						if (xCurent == 3) {
							letsBEEP = 1;
						}
						//cli();
						move = 2;
					}
					break;

				case 0b11110111: /* top_button */
					if (!gameEnded) {
						if (yCurent == 1) {
							letsBEEP = 1;
						}
						//cli();
						move = 1;
					}
					break;

				case 0b11101111: /* ack_button */
					if (!gameEnded) {
						if (m[yCurent][xCurent] == 0) { /* selectia este libera */
							cli();
							turnLEDon(yCurent, xCurent, culori[playerCurent], 0); /* aprind ledul selectat */
							nrLEDuriAprinse++;
							if (aCastigat())
								gameEnded = 1;
							if (nrLEDuriAprinse == 9)
								gameEnded = 1;
							if (!gameEnded) {
								INC(playerCurent); /* schimb playerul */
								turnLEDon(0, 0, culori[playerCurent], 0); /* schimb indicatorul de stare */
								mutaUrmatoareaPozitie = 1;
							}
							else {
								sei();
							}
						}
						else {
							letsBEEP = 1;
						}
					}
					else {
						//todo
						//restart game
						//reset();
					}
					break;

				case 0b11011111: /* left_button */
					if (!gameEnded) {
						if (xCurent == 1) {
							letsBEEP = 1;
						}
						//cli();
						move = 4;
					}
					break;

				case 0b10111111: /* bottom_button */
					if (!gameEnded) {
						if (yCurent == 3) {
							letsBEEP = 1;
						}
						//cli();
						move = 3;
					}
					break;

				default:
					if (!gameEnded) { /* nu stiu de ce cateodata butonul de jos emite altceva decat 0b10111111 */
						if (yCurent == 3) {
							letsBEEP = 1;
						}
						//cli();
						move = 3;
					}
					break;
			}
		} else {
			if (mutaUrmatoareaPozitie) {
				for (i = 1; i < 4; i++) {
					for (j = 1; j < 4; j++)
						if (m[i][j] == 0) {
							xCurent = j;
							yCurent = i;
							ok = 1;
							break;
						}
					if (ok) {
						ok = 0;
						break;
					}
				}
				mutaUrmatoareaPozitie = 0;
				sei();
			}
			if (move) {
				switch (move) {
					case 1:
						if (yCurent != 1) {
							turnLEDoff(yCurent, xCurent, 1);
							if (m[yCurent][xCurent])
								turnLEDon(yCurent, xCurent, culori[m[yCurent][xCurent]], 0);
							yCurent--;
						}
					break;
					case 2:
						if (xCurent != 3) {
							turnLEDoff(yCurent, xCurent, 1);
							if (m[yCurent][xCurent])
								turnLEDon(yCurent, xCurent, culori[m[yCurent][xCurent]], 0);
							xCurent++;
						}
					break;
					case 3:
						if (yCurent != 3) {
							turnLEDoff(yCurent, xCurent, 1);
							if (m[yCurent][xCurent])
								turnLEDon(yCurent, xCurent, culori[m[yCurent][xCurent]], 0);
							yCurent++;
						}
					break;
					case 4:
						if (xCurent != 1) {
							turnLEDoff(yCurent, xCurent, 1);
							if (m[yCurent][xCurent])
								turnLEDon(yCurent, xCurent, culori[m[yCurent][xCurent]], 0);
							xCurent--;
						}
					break;
				}
				move = 0;
				//sei();
			}
		}
		displayLEDs(0);
	}

	return 0;
}